// Assignment 1
// CSC 445 - Lea
// Jacob M. Peck

// network test app
// run as server: java NetTestApp Server
// run as client: java NetTestApp Client <serverhost>

// To Do: fix Connection Reset on throughout measurements

import java.util.*;
import java.net.*;
import java.io.*;

public class NetTestApp {
  private static final int NUMTESTS = 100;
  private static final int portnum = 2700;
  private static boolean server = false;
  private static String serverhost;
  private static ServerSocket ss;
  
  private static final int S_1 = 1;
  private static final int S_4 = 4;
  private static final int S_16 = 16;
  private static final int S_256 = 256;
  
  private static final int S_1K = 1024;
  private static final int S_4K = 4096;
  private static final int S_16K = 16384;
  private static final int S_64K = 65536;
  private static final int S_256K = 262114;
  private static final int S_1M = 1048576;
  
  private static final int S_2K = 2048;
  private static final int S_512 = 512;
  private static final int S_128 = 128;
  private static final int S_64 = 64;
  
  public static void main(String[] args) {
    if (args.length < 1) {
      System.out.println("Usage: either java NetTestApp Server or java NetTestApp Client <ServerHost>");
      System.exit(-1);
    }
    if (args[0].equals("Server")) server = true;
    
    if (server)
      System.out.println("Running in SERVER mode on port " + portnum);
    else {
      serverhost = args[1];
      System.out.println("Running in CLIENT mode, connecting to " + serverhost + " on port " + portnum);
    }
        
    try {
      if(server) ss = new ServerSocket(portnum);
      //Thread.sleep(1000);
      System.out.println("UDP Section\n-----------");
      System.out.println("UDP RTTs --");
      testUDPMessageSize();
      
      //Thread.sleep(1000);
      System.out.println("\nUDP Throughput --");
      testUDPThroughput();
      
      
      //Thread.sleep(1000);
      System.out.println("\n\nTCP Section\n-----------");
      System.out.println("TCP RTTs --");
      testTCPMessageSize();
      
      //Thread.sleep(1000);
      System.out.println("\nTCP Throughput --");
      testTCPThroughput();
      
      //Thread.sleep(1000);
      System.out.println("\nTCP Message Interactions --");
      testTCPMessageInteraction();
      
    } catch (Exception ex) {ex.printStackTrace();}
    
  }
  
  // UDP Message Size
  private static void testUDPMessageSize() throws Exception {
    if(server) {
      testUDPMessageSizeServer();
      return;
    }
    // make connection
    DatagramSocket sock = new DatagramSocket();
    InetAddress host = InetAddress.getByName(serverhost);
    sock.connect(host, portnum);
    
    Thread.sleep(100);
    float total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += UDPSend(S_1, sock, host);
    System.out.println("  UDP RTT (1): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += UDPSend(S_4, sock, host);
    System.out.println("  UDP RTT (4): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += UDPSend(S_16, sock, host);
    System.out.println("  UDP RTT (16): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += UDPSend(S_256, sock, host);
    System.out.println("  UDP RTT (256): " + total / NUMTESTS + "ms");
    
    sock.close();
  }
  
  private static float UDPSend(int size, DatagramSocket sock, InetAddress host) throws Exception {
    // send packet
    int numtimes = 1;
    long totaltime = 0;
    
    if (size > S_16K) {
      numtimes = size / S_16K;
    }
    
    DatagramPacket dp;
    for(int i = 0; i < numtimes; i++) {
      if(numtimes > 1)
        dp = new DatagramPacket(new byte[S_16K], S_16K, host, portnum);
      else
        dp = new DatagramPacket(new byte[size], size, host, portnum);
      long start = System.nanoTime();
      sock.send(dp);
      
      // receive packet
      sock.receive(dp);
      long end = System.nanoTime();
      totaltime += (end - start);
    }
    return totaltime/1000000f;
  }
  
  private static void testUDPMessageSizeServer() throws Exception {
    // make connection
    DatagramSocket sock = new DatagramSocket(portnum);
    
    for(int i = 0; i < NUMTESTS; i++)
      UDPRecv(S_1, sock);
    for(int i = 0; i < NUMTESTS; i++)
      UDPRecv(S_4, sock);
    for(int i = 0; i < NUMTESTS; i++)
      UDPRecv(S_16, sock);
    for(int i = 0; i < NUMTESTS; i++)
      UDPRecv(S_256, sock);
    
    sock.close();
  }
  
  private static void UDPRecv(int size, DatagramSocket sock) throws Exception {
    // receive packet
    int numtimes = 1;
    if(size > S_16K)
      numtimes = size / S_16K;
    
    
    DatagramPacket dp;
    
    for(int i = 0; i < numtimes; i++) {
      if(numtimes > 1)
        dp = new DatagramPacket(new byte[S_16K], S_16K);
      else
        dp = new DatagramPacket(new byte[size], size);
      sock.receive(dp);
      
      // send packet
      sock.send(dp);
    }
    System.out.println("  UDP: Sent back message of size " + size);
  }
  
  // UDP Throughput
  private static void testUDPThroughput() throws Exception {
    if(server) {
      testUDPThroughputServer();
      return;
    }
    // make connection
    DatagramSocket sock = new DatagramSocket();
    InetAddress host = InetAddress.getByName(serverhost);
    sock.connect(host, portnum);
    
    Thread.sleep(100);
    System.out.println("  UDP Throughput (1024): " + calculateThroughput(S_1K, UDPSend(S_1K, sock, host)) + "Mbps");
    System.out.println("  UDP Throughput (4096): " + calculateThroughput(S_4K, UDPSend(S_4K, sock, host)) + "Mbps");
    System.out.println("  UDP Throughput (16384): " + calculateThroughput(S_16K, UDPSend(S_16K, sock, host)) + "Mbps");
    System.out.println("  UDP Throughput (65536): " + calculateThroughput(S_64K, UDPSend(S_64K, sock, host)) + "Mbps");
    System.out.println("  UDP Throughput (262144): " + calculateThroughput(S_256K, UDPSend(S_256K, sock, host)) + "Mbps");
    System.out.println("  UDP Throughput (1048576): " + calculateThroughput(S_1M, UDPSend(S_1M, sock, host)) + "Mbps");
    
    sock.close();
  }
  
  private static double calculateThroughput(int size, float rtt) {
    float time = rtt / 2f; // one-way time
    long bits = size * 8; // 8 bits per byte
    float numerator = bits * 1000;
    float denominator = time * 1000000;
    return numerator/denominator;
  }
  
  private static void testUDPThroughputServer() throws Exception {
    // make connection
    DatagramSocket sock = new DatagramSocket(portnum);
    
    UDPRecv(S_1K, sock);
    UDPRecv(S_4K, sock);
    UDPRecv(S_16K, sock);
    UDPRecv(S_64K, sock);
    UDPRecv(S_256K, sock);
    UDPRecv(S_1M, sock);
    
    sock.close();  
  }
  
  // TCP Message Size
  private static void testTCPMessageSize() throws Exception {
    if(server) {
      testTCPMessageSizeServer();
      return;
    }
    // make connection
    Socket sock = new Socket(serverhost, portnum);
    
    Thread.sleep(100);
    float total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += TCPSend(S_1, sock);
    System.out.println("  TCP RTT (1): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += TCPSend(S_4, sock);
    System.out.println("  TCP RTT (4): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += TCPSend(S_16, sock);
    System.out.println("  TCP RTT (16): " + total / NUMTESTS + "ms");
    total = 0;
    for(int i = 0; i < NUMTESTS; i++)
      total += TCPSend(S_256, sock);
    System.out.println("  TCP RTT (256): " + total / NUMTESTS + "ms");
    
    sock.close();
  }
  
  private static float TCPSend(int size, Socket sock) throws Exception {
    // send packet
    OutputStream out = sock.getOutputStream();
    InputStream in = sock.getInputStream();
    DataInputStream di = new DataInputStream(in);
    int numtimes = 1;
    long totaltime = 0;
    byte[] data;
    if (size > S_16K)
      numtimes = size / S_16K;
    for(int i = 0; i < numtimes; i++) {
      //System.out.println(i);
      if (numtimes > 1)
        data = new byte[S_16K];
      else
        data = new byte[size];
      long start = System.nanoTime();
      //System.out.println("write.");
      out.write(data);
      //System.out.println("sent. " +  size);
      // receive packet
      //System.out.println("read.");
      di.readFully(data);
      //System.out.println("received. " + size);
      long end = System.nanoTime();
      totaltime += (end - start);
    }
    return totaltime/1000000f;
  }
  
  private static void testTCPMessageSizeServer() throws Exception {
    // make connection
    Socket sock = ss.accept();
    
    for(int i = 0; i < NUMTESTS; i++)
      TCPRecv(S_1, sock);
    for(int i = 0; i < NUMTESTS; i++)
      TCPRecv(S_4, sock);
    for(int i = 0; i < NUMTESTS; i++)
      TCPRecv(S_16, sock);
    for(int i = 0; i < NUMTESTS; i++)
      TCPRecv(S_256, sock);
    
    //sock.close();
  }
  
  private static void TCPRecv(int size, Socket sock) throws Exception {
    // receive packet
    int numtimes = 1;
    byte[] data;
    if(size > S_16K)
      numtimes = size / S_16K;
    //System.out.println("Socket: " + sock + " Size: " + size);
    OutputStream out = sock.getOutputStream();
    InputStream in = sock.getInputStream();
    DataInputStream di = new DataInputStream(in);
    for(int i = 0; i < numtimes; i++) {
      //System.out.println(i);
      if(numtimes > 1)
        data = new byte[S_16K];
      else
        data = new byte[size];
      di.readFully(data);
      
      // send packet
      out.write(data);
    }
    System.out.println("  TCP: Sent back message of size " + size);
  }
  
  // TCP Throughput
  private static void testTCPThroughput() throws Exception {
    if(server) {
      testTCPThroughputServer();
      return;
    }
    // make connection
    Socket sock = new Socket(serverhost, portnum);
    
    Thread.sleep(100);
    System.out.println("  TCP Throughput (1024): " + calculateThroughput(S_1K, TCPSend(S_1K, sock)) + "Mbps");
    System.out.println("  TCP Throughput (4096): " + calculateThroughput(S_4K, TCPSend(S_4K, sock)) + "Mbps");
    System.out.println("  TCP Throughput (16384): " + calculateThroughput(S_16K, TCPSend(S_16K, sock)) + "Mbps");
    System.out.println("  TCP Throughput (65536): " + calculateThroughput(S_64K, TCPSend(S_64K, sock)) + "Mbps");
    System.out.println("  TCP Throughput (262144): " + calculateThroughput(S_256K, TCPSend(S_256K, sock)) + "Mbps");
    System.out.println("  TCP Throughput (1048576): " + calculateThroughput(S_1M, TCPSend(S_1M, sock)) + "Mbps");
    //System.out.println("  TCP Throughput (524288): " + (524288 / ((TCPSend(524288, sock))/2000000000f)) + "bps");
    
    sock.close();
  }
  
  private static void testTCPThroughputServer() throws Exception {
    Socket sock = ss.accept();
    
    TCPRecv(S_1K, sock);
    TCPRecv(S_4K, sock);
    TCPRecv(S_16K, sock);
    TCPRecv(S_64K, sock);
    TCPRecv(S_256K, sock);
    TCPRecv(S_1M, sock);
    //TCPRecv(524289, sock);
    
    sock.close();
  }
  
  // TCP Message Interaction
  private static void testTCPMessageInteraction() throws Exception {
    if(server) {
      testTCPMessageInteractionServer();
      return;
    }
    
    Socket sock = new Socket(serverhost, portnum);
    Thread.sleep(100);
    System.out.println("  TCP Interaction (256x4K): " + TCPinteractionSend(256, S_4K, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (512x2K): " + TCPinteractionSend(512, S_2K, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (1024x1K): " + TCPinteractionSend(1024, S_1K, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (2048x512): " + TCPinteractionSend(2048, S_512, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (4096x256): " + TCPinteractionSend(4096, S_256, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (8192x128): " + TCPinteractionSend(8192, S_128, sock) + "ms");
    Thread.sleep(100);
    System.out.println("  TCP Interaction (16384x64): " + TCPinteractionSend(16384, S_64, sock) + "ms");
    sock.close();
  }
  
  private static float TCPinteractionSend(int num, int size, Socket sock) throws Exception {
    //System.out.println(size);
    OutputStream out = sock.getOutputStream();
    InputStream in = sock.getInputStream();
    DataInputStream di = new DataInputStream(in);
    byte[] data;
    
    long start = System.nanoTime();
    for(int i = 0; i < num; i++) {
      //System.out.println(i);
      data = new byte[size];
      out.write(data);
    }
    //System.out.println("Wrote.");
    di.readFully(new byte[S_1]);
    //System.out.println("Read.");
    long end = System.nanoTime();
    return (end - start)/1000000f;
  }
  
  private static void testTCPMessageInteractionServer() throws Exception {
    Socket sock = ss.accept();
    
    TCPinteractionRecv(256, S_4K, sock);
    TCPinteractionRecv(512, S_2K, sock);
    TCPinteractionRecv(1024, S_1K, sock);
    TCPinteractionRecv(2048, S_512, sock);
    TCPinteractionRecv(4096, S_256, sock);
    TCPinteractionRecv(8192, S_128, sock);
    TCPinteractionRecv(16384, S_64, sock);
    Thread.sleep(1000);
    sock.close();
  }
  
  private static void TCPinteractionRecv(int num, int size, Socket sock) throws Exception {
    //System.out.println(size);
    OutputStream out = sock.getOutputStream();
    InputStream in = sock.getInputStream();
    DataInputStream di = new DataInputStream(in);
    byte[] data;

    for(int i = 0; i < num; i++) {
      //System.out.println(i);
      data = new byte[size];
      di.readFully(data);
    }
    //System.out.println("Read.");
    out.write(new byte[S_1]);
    System.out.println("  Sent back an ack for (" + num + "x" + size + ")");
  }
}
